//+------------------------------------------------------------------+
//|                                            Copyist positions.mq4 |
//|                                                            Urain |
//|                               http://www.mql4.com/ru/users/Urain |
//+------------------------------------------------------------------+
#property copyright "Urain"
#property link      "http://www.mql4.com/ru/users/Urain"
#property show_inputs
extern double koeff_lot=1.0; // position volume coefficient
int delay_time=500;  // delay time
int  cnt_command=0;  // command counter
int  TotalSymbols;   // total symbols
string Symbols[];    // symbol names
double SymPosVol[];  // symbol volumes
int    SymPosType[]; // symbol position types (-1 if no position)
double SymPosSL[];   // symbol position Stop Loss price
double SymPosTP[];   // symbol position Take Profit price
//--- real volumes
double realSymPosVol[];  // current value of position volume
int    realSymPosType[]; // current value of position type (or -1 if there isn't position)
double realSymPosSL[];   // current value of position Stop Loss
double realSymPosTP[];   // current value of position Take Profit
int Size=0; // number of lines in the file, also used in resize of the arrays
string s[]; // array for file lines
//+------------------------------------------------------------------+
//|  Script program start function                                   |
//+------------------------------------------------------------------+
int start()
  {
//--- read the current position states from file
   Size=READS("Translator positions",s,true);
   ArrayResize(Symbols,Size);
//--- fill array with symbol names
   for(int i=1;i<Size;i++)parser_string_Symbols(i);
   string com0="";// comment
   for(i=1;i<Size;i++)// find symbols without data
     {
      if(bid(Symbols[i])==0)
        {
         //--- if symbol is empty, print it in comment
         com0=com0+"["+Symbols[i]+"] ";
         Symbols[i]="";
        }
     }
//--- print comment on symbols without data
   if(com0!="")com0="Symbols without data "+com0+"\n";

//--- set size of the buffers
   ArrayResize(SymPosVol,Size);
   ArrayResize(SymPosType,Size);
   ArrayResize(SymPosSL,Size);
   ArrayResize(SymPosTP,Size);

   ArrayResize(realSymPosVol,Size);
   ArrayResize(realSymPosType,Size);
   ArrayResize(realSymPosSL,Size);
   ArrayResize(realSymPosTP,Size);

//--- infinite loop, it will finish when program will be terminated by user
   while(!IsStopped())
     {
      if(Busy_and_Connected())// if trade context is not busy
        {
         READS("Translator positions",s,false);
         if(cnt_command!=StrToInteger(s[0]))
           {
            //--- parse file and get size
            cnt_command=parser(Size);
            string com=com0+s[0];
            //---  prepare comment line
            for( i=1;i<Size;i++)
               com=com+"\n"+Symbols[i]+"  [ "+DoubleToStr(SymPosVol[i],DigitsMinLot(Symbols[i]))+" ]  [ "+SymPosType[i]+" ]  [ "+
                   DoubleToStr(SymPosSL[i],digits(Symbols[i]))+" ]  [ "+
                   DoubleToStr(SymPosTP[i],digits(Symbols[i]))+" ]";
            //--- print comment
            Comment(com);
           }
         //--- save start time
         int start=GetTickCount();
         //--- if position volumes are not equal, process the signals
         if(compare_positions())processing_signals();
         //--- if SL/TP levels has changed, proceed them
         if(compare_sl_tp_levels())processing_sl_tp_levels();
         //--- calculate time
         int TickCount=GetTickCount()-start;
        }
      //--- sleep the remained time
      if(delay_time>TickCount)Sleep(delay_time-TickCount);
     }
//--- print alert with information about end of the work
   Alert("end, Copyist positions released");
//--- clear comments
   Comment("");
   return(0);
  }
//+------------------------------------------------------------------+
//| CUSTOM SERVICE FUNCTIONS                                         |
//+------------------------------------------------------------------+  
//--- digits on the symbol
int digits(string symbol){return(MarketInfo(symbol,MODE_DIGITS));}
//--- point size
double point(string symbol){return(MarketInfo(symbol,MODE_POINT));}
//--- ask price
double ask(string symbol){return(MarketInfo(symbol,MODE_ASK));}
//--- bid price
double bid(string symbol){return(MarketInfo(symbol,MODE_BID));}
//--- spread 
int spred(string symbol){return(MarketInfo(symbol,MODE_SPREAD));}
//--- stop level
int stlevel(string symbol){return(MarketInfo(symbol,MODE_STOPLEVEL));}
//--- max lot 
double maxlot(string symbol){return(MarketInfo(symbol,MODE_MAXLOT));}
//--- min lot
double minlot(string symbol){return(MarketInfo(symbol,MODE_MINLOT));}
//--- normalize v (price)
double nd(string symbol,double v){return(NormalizeDouble(v,digits(symbol)));}
//--- normalize v (step size)
double NDL(string symbol,double v){return(NormalizeDouble(v,DigitsMinLot(symbol)));}
//--- open price 
double price_open(string symbol,int cmd) {if(cmd%2)return(bid(symbol));else return(nd(symbol,ask(symbol)));}
//--- close price
double price_close(string symbol,int cmd){if(cmd%2)return(ask(symbol));else return(nd(symbol,bid(symbol)));}
//--- check connection state
bool Busy_and_Connected(){if(IsTradeAllowed() && IsConnected())return(true);else return(false);}
//+------------------------------------------------------------------+
//| Lot step on the symbol                                           |
//+------------------------------------------------------------------+
int DigitsMinLot(string symbol)
  {
   double ml=MarketInfo(symbol,MODE_MINLOT);
//--- 1/x of lot step
   double Dig=0;if(ml!=0)Dig=1.0/ml;
//--- conversion of 1/x to digits
   double res=0;
   if(Dig>1)res=1;
   if(Dig>10)res=2;
   if(Dig>100)res=3;
   if(Dig>1000)res=4;
   return(res);
  }
//+------------------------------------------------------------------+
//| Normalize lot to min/max lot on the symbol                       |
//| and return the normalized lot (lot step)                         | 
//+------------------------------------------------------------------+
double ndlot(string symbol,double l)
  {
   double ml=minlot(symbol);
   double rez=NormalizeDouble(l/ml,0)*ml;
   double Maxlot=maxlot(symbol);
   if(rez>Maxlot)rez=Maxlot;
   return(NormalizeDouble(rez,DigitsMinLot(symbol)));
  }
//+------------------------------------------------------------------+
//| Positions processing functions                                   |
//+------------------------------------------------------------------+
//+------------------------------------------------------------------+
//| Compare positions                                                |
//+------------------------------------------------------------------+
bool compare_positions()
  {
//--- get position volumes on all symbols
   real_pos_volum();
//--- loop
   for(int i=1;i<Size;i++)
     {
      //--- if there isn't a symbol, skip iteration
      if(Symbols[i]=="")continue;
      //--- compare the normalized real volumes and volumes in the file, if they aren't equal, exit
      if(NDL(Symbols[i],realSymPosVol[i])!=NDL(Symbols[i],SymPosVol[i])){return(true);}
      //--- compare the real position types and types in the file, if they aren't equal, exit
      if(realSymPosType[i]!=SymPosType[i]){return(true);}
     }
   return(false);
  }
//+------------------------------------------------------------------+
//| Get real position volume on all symbols                          |
//+------------------------------------------------------------------+
void real_pos_volum()
  {
//--- initialize arrays
   ArrayInitialize(realSymPosVol,0.0);
   ArrayInitialize(realSymPosType,-1);
//--- get number of total orders
   int Total=OrdersTotal();
//--- loop all orders
   for(int i=0;i<Total;i++)
     {
      //--- select the order
      if(OrderSelect(i,SELECT_BY_POS))
        {
         //--- save order type in a variable
         int Type=OrderType();
         //--- if market order
         if(Type%2==Type)
           {
            //--- save symbol in a variable
            string symbol=OrderSymbol();
            //--- loop to compare with file lines
            for(int j=1;j<Size;j++)
              {
               //--- if we have this order symbol 
               if(symbol==Symbols[j])
                 {
                  //--- perform summation of order volumes into the total net position
                  realSymPosVol[j]+=NDL(symbol,OrderLots()*(Type*2-1));
                 }
              }
           }
        }
     }
//--- loop all the symbols
   for(i=1;i<Size;i++)
     {
      //--- if symbol is empty, skip iteration
      if(Symbols[i]=="")continue;
      //--- procced if symbol position volume isn't equal to 0
      if(realSymPosVol[i]!=0)
        {
         //--- real position type
         realSymPosType[i]=(realSymPosVol[i]/MathAbs(realSymPosVol[i])+1)/2;
         //--- real position volume
         realSymPosVol[i]=MathAbs(realSymPosVol[i]);
        }
     }
   return;
  }
//+------------------------------------------------------------------+
//| Proceed signals                                                  |
//+------------------------------------------------------------------+
void processing_signals()
  {
   while(!IsStopped())
     {
      //--- if trade context is not busy
      if(Busy_and_Connected())
        {
         //--- loop all symbols
         for(int i=1;i<Size;i++)
           {
            //--- if symbol is empty, continue
            if(Symbols[i]=="")continue;
            //--- convert position direction to form: -1,+1            
            int TF=SymPosType[i]*2-1;
            //--- convert the real position type to form -1,+1
            int TR=realSymPosType[i]*2-1;
            //--- save position volume (from file)
            double VF=SymPosVol[i];
            //--- save position volume (real volume)
            double VR=realSymPosVol[i];
            double lot;
            //--- if positions aren't equal
            if(NormalizeDouble(VF*TF,8)!=NormalizeDouble(VR*TR,8))
              {
               //--- if real volume isn't equal to zero and directions aren't equal or
               //--- if the directions are equal but the real volume is greater than specified in file
               if((VR!=0 && TF!=TR) || (TF==TR && VF<VR))
                 {
                  //--- directions are the same, but real volume is greater than specified in file
                  if(TF==TR && VF<VR)lot=realSymPosVol[i]-SymPosVol[i];
                  //--- if the real volume isn't equal to 0, and directions are different
                  else lot=realSymPosVol[i];
                  //--- close the volume and break
                  close_market_order(Symbols[i],lot);
                  break;
                 }
               else
                 {
                  //--- if directions are equal and real volume is lower than specified in file
                  if(TF==TR && VF>VR)lot=SymPosVol[i]-realSymPosVol[i];
                  //--- if directions aren't equal and real volume is equal to zero
                  else lot=SymPosVol[i];
                  //--- open market order and break
                  open_market_order(Symbols[i],SymPosType[i],lot);
                  break;
                 }
              }
            else
              {
               //--- if positions are equal and trade context is busy, break
               if(!IsTradeAllowed())break;
              }
           }
        }
      //--- if trade context is busy, show comment
      else Comment("No connection or trade context is busy");
      //--- if positions are equal, break
      if(!compare_positions())break;
     }
   return;
  }  
//+------------------------------------------------------------------+
//| Close market order                                               |
//+------------------------------------------------------------------+
void close_market_order(string symbol,double lot)
  {
//--- order close flag
   bool FlagLot=true;
//--- get number of trade orders
   int Total=OrdersTotal();
//--- loop all orders
   for(int i=0;i<Total;i++)
     {
      //--- select order
      if(OrderSelect(i,SELECT_BY_POS))
        {
         //--- if market order
         if(OrderType()%2==OrderType())
           {
            //--- compare symbol with order symbol
            if(symbol==OrderSymbol())
              {
               //--- check if the close volume is sufficient to close all
               if(OrderLots()==ndlot(symbol,lot))
                 {
                  //--- close order
                  Closes(symbol,OrderTicket(),lot);
                  //--- set flag to false and break
                  FlagLot=false;
                  break;
                 }
              }
           }
        }
     }
//---
//---
   if(FlagLot)
     {
      //--- loop all orders
      for(i=0;i<Total;i++)
        {
         //--- select the order
         if(OrderSelect(i,SELECT_BY_POS))
           {
            //--- if order type isn't pending order
            if(OrderType()%2==OrderType())
              {
               //--- compare order symbol
               if(symbol==OrderSymbol())
                 {
                  //--- close order with specified volume and break
                  Closes(symbol,OrderTicket(),lot);
                  break;
                 }
              }
           }
        }
     }
   return;
  }
//+------------------------------------------------------------------+
//| Stop Loss and Take Profit levels processsing                     |
//+------------------------------------------------------------------+
//+------------------------------------------------------------------+
//| compare SL/TP levels                                             |
//+------------------------------------------------------------------+
bool compare_sl_tp_levels()
  {
//--- get current stop levels
   real_pos_sl_tp_levels();
//--- compare stop levels
   for(int i=1;i<Size;i++)
     {
      //--- if symbol is empty, continue
      if(Symbols[i]=="")continue;
      //--- compare SL, if they aren't equal, exit
      if(nd(Symbols[i],realSymPosSL[i])!=nd(Symbols[i],SymPosSL[i])){return(true);}
      //--- compare TP, if they aren't equal, exit
      if(nd(Symbols[i],realSymPosTP[i])!=nd(Symbols[i],SymPosTP[i])){return(true);}
     }
//--- 
   return(false);
  }
//+------------------------------------------------------------------+
//| get info about current stop levels                               |
//+------------------------------------------------------------------+
void real_pos_sl_tp_levels()
  {
//--- initialize arrays with 0
   ArrayInitialize(realSymPosSL,0.0);
   ArrayInitialize(realSymPosTP,0.0);
//--- get total orders
   int Total=OrdersTotal();
//--- loop all orders
   for(int i=0;i<Total;i++)
     {
      //--- select order
      if(OrderSelect(i,SELECT_BY_POS))
        {
         //--- order direction
         int Type=OrderType();
         //--- market order
         if(Type%2==Type)
           {
            //--- order symbol
            string symbol=OrderSymbol();
            //--- compare with lines from the file
            for(int j=1;j<Size;j++)
              {
               //--- symbol
               if(symbol==Symbols[j])
                 {
                  //--- if not equal, set -EMPTY_VALUE                  
                  if(NormalizeDouble(OrderStopLoss(),digits(symbol))!=SymPosSL[j])
                     realSymPosSL[j]=-EMPTY_VALUE;
                  else
                    {
                     //--- if it isn't equal to -EMPTY_VALUE, set the value from file
                     if(realSymPosSL[j]!=-EMPTY_VALUE)realSymPosSL[j]=SymPosSL[j];
                    }
                  //--- if value isn't equal, set -EMPTY_VALUE  
                  if(NormalizeDouble(OrderTakeProfit(),digits(symbol))!=SymPosTP[j])
                     realSymPosTP[j]=-EMPTY_VALUE;
                  else
                    {
                     //--- if it isn't equal to -EMPTY_VALUE, set the value from file
                     if(realSymPosTP[j]!=-EMPTY_VALUE)realSymPosTP[j]=SymPosTP[j];
                    }
                 }
              }
           }
        }
     }
   return;
  }
//+------------------------------------------------------------------+
//| Select order and modify stop levels                              |
//+------------------------------------------------------------------+
void modification_sl_tp_levels()
  {
//--- get total number of orders
   int Total=OrdersTotal();
//--- loop all orders
   for(int i=0;i<Total;i++)
     {
      //--- select order
      if(OrderSelect(i,SELECT_BY_POS))
        {
         //--- order direction
         int Type=OrderType();
         //--- if market order
         if(Type%2==Type)
           {
            //--- order symbol
            string symbol=OrderSymbol();
            //--- loop all lines from the file
            for(int j=1;j<Size;j++)
              {
               //--- if order symbol is equal to the line from the file
               if(symbol==Symbols[j])
                 {
                  //--- if the real SL/TP aren't equal to the values from the file
                  if(NormalizeDouble(OrderStopLoss(),digits(symbol))!=SymPosSL[j] || 
                     NormalizeDouble(OrderTakeProfit(),digits(symbol))!=SymPosTP[j]
                     )
                    {
                     //--- modify stop levels and break
                     OrderTradeModif(OrderTicket(),symbol,Type,OrderOpenPrice(),SymPosSL[j],SymPosTP[j]);
                     break;
                    }
                 }
              }
           }
        }
     }
  }  
//+------------------------------------------------------------------+
//| processing of sl and tp levels                                   |
//+------------------------------------------------------------------+
void processing_sl_tp_levels()
  {
//--- start time
   int start=GetTickCount();
   while(!IsStopped())
     {
      //--- if trade context is not busy
      if(Busy_and_Connected())
        {
         //--- select order and modify SL and TP levels
         modification_sl_tp_levels();
        }
      //--- update information from file if the time is greater than delay time
      if(GetTickCount()-start>delay_time)READS("Translator positions",s,false);
      //--- break if command counter (in the file) has changed
      if(cnt_command!=StrToInteger(s[0]))break;
      //--- small sleep
      Sleep(50);
      //--- if the real stop levels are equal to stop levels in file, break
      if(!compare_sl_tp_levels())break;
     }
   return;
  }
//+------------------------------------------------------------------+
//| Working with orders                                              |
//+------------------------------------------------------------------+  
//+------------------------------------------------------------------+
//| Close or decrease the volume of order with the specified ticket  |
//+------------------------------------------------------------------+
bool Closes(string symbol,int ticket,double lot)
  {
//--- get order volume
   double Lots=OrderLots();
//--- if the volume is equal to the order volume, normalize volume
   if(Lots>lot)Lots=ndlot(symbol,lot);
//--- if the order is order of pending type, or normalized volume is less or equal to 0, return
   if(OrderType()>1 || Lots<=0)return;
//--- if all checks ok, close the order
   if(!OrderClose(ticket,Lots,price_close(OrderSymbol(),OrderType()),spred(OrderSymbol())))
     {
      //--- if error, returns false and show error code
      Print("Error in order closing() ",GetLastError());
      return(false);
     }
   return(true); // if ok, return true
  }
//+------------------------------------------------------------------+
//| Modify stop levels of marker order                               |
//+------------------------------------------------------------------+
bool OrderTradeModif(int ticket,string symbol,int cmd,double price,double stoploss=0,double takeprofit=0,int magic=0)
  {
//--- local variables
   double Stoploss,Takeprofit;
//---  flag to disable modify because of conditions
   bool flag=false;

//---  check stop loss
   if(stoploss!=0)
     {
      //--- if stop loss and take profit prices are wrong, swap them
      if((stoploss>price && !cmd) || 
         (stoploss<price && cmd)
         )Stoploss=takeprofit; else Stoploss=stoploss;
      //--- check if stop loss price is too close to the market
      if(MathAbs(price_close(symbol,cmd)-Stoploss)<stlevel(symbol)*point(symbol))
         Stoploss=OrderStopLoss();
     }
   if(takeprofit!=0)
     {
      //--- if stop loss and take profit prices are wrong, swap them
      if((takeprofit<price && !cmd) || 
         (takeprofit>price && cmd)
         )Takeprofit=stoploss; else Takeprofit=takeprofit;
      //--- check if take profit price is too close to the market
      if(MathAbs(price_close(symbol,cmd)-Takeprofit)<stlevel(symbol)*point(symbol))
         Takeprofit=OrderTakeProfit();
     }
//--- is the modification needed?
   if(Stoploss==OrderStopLoss() && Takeprofit==OrderTakeProfit())return(false);
   else return(OrderModify(ticket,price,Stoploss,Takeprofit,NULL));
  }
//+------------------------------------------------------------------+
//| Open market order                                                |
//+------------------------------------------------------------------+
int open_market_order(string symbol,int cmd,double volume,int stoploss=0,int takeprofit=0,int magic=0)
  {
//--- local variables
   int direction=((cmd%2)*2)-1,Cmd=cmd;
   double Stoploss=0,Takeprofit=0,Price;

//--- open price calculation
   if(cmd<2)// market order
      Price=price_open(symbol,cmd);

//--- checks to set SL/TP with the order
/*   if(stoploss!=0 && stoploss<stlevel(symbol))Stoploss=stlevel(symbol);
   if(takeprofit!=0 && takeprofit<stlevel(symbol))Takeprofit=stlevel(symbol);
   
   if(stoploss!=0)
      Stoploss=nd(symbol,Price+Stoploss*point(symbol)*direction);
   else Stoploss=0;

   if(takeprofit!=0)
      Takeprofit=nd(symbol,Price-Takeprofit*point(symbol)*direction);
   else Takeprofit=0; */

//--- normalize lot
   double lot=ndlot(symbol,volume);
   if(lot==0)return(-1);

//--- send order
   int ticket=OrderSend(symbol,Cmd,lot,Price,spred(symbol),Stoploss,Takeprofit,NULL,magic);

//--- check order ticket
   if(ticket>-1)return(ticket);
   else Print("Error opening order() ",GetLastError());
  }
//+------------------------------------------------------------------+
//| Function for working with file                                   |
//+------------------------------------------------------------------+
//+------------------------------------------------------------------+
//| read position states from the file                               |
//+------------------------------------------------------------------+
int READS(string files,string &s[],bool resize)
  {
//--- open file for reading
   int handle=FileOpen(files+".csv",FILE_CSV|FILE_READ,";");
//--- if successful
   if(handle>0)
     {
      //--- set initial array size, if it first pass
      if(resize)ArrayResize(s,1);
      int cnt=0;
      while(true)// loop until break
        {
         //--- increase array size
         if(resize)ArrayResize(s,cnt+1);
         s[cnt]=FileReadString(handle);  // read string from the file
         FileSeek(handle,0,SEEK_CUR);    // seek at new position
         if(s[cnt]=="")break; // break
         cnt++;
        }
      ArrayResize(s,cnt);
      FileClose(handle);
     }
//--- if error
   else Print("Error opening file ",GetLastError());
   return(cnt);// returns the number of lines read
  }
//+------------------------------------------------------------------+
//| Parse file                                                       |
//+------------------------------------------------------------------+
int parser(int Size)
  {
//--- parse file lunes
   for(int i=1;i<Size;i++)
      parser_string(i);
//--- returns the counter
   return(StrToInteger(s[0]));
  }
//+------------------------------------------------------------------+
//| Parse symbol name                                                |
//+------------------------------------------------------------------+
void parser_string_Symbols(int x)
  {
//--- get line length
   int Len=StringLen(s[x]);
//--- starting index
   int start=0;
//--- find position of ","   
   int end=StringFind(s[x],",",start);
//--- calculate line length
   int length=end-start;
//--- get substring with symbol
   Symbols[x]=StringSubstr(s[x],start,length);
   return;
  }
//+------------------------------------------------------------------+
//| Parse the other line elements                                    |
//+------------------------------------------------------------------+
void parser_string(int x)
  {
//--- get line length
   int Len=StringLen(s[x]);
//--- starting position
   int start=7;
//--- find position of ","
   int end=StringFind(s[x],",",start);
//--- calculate the length of the substring
   int length=end-start;
//--- get Volume
   SymPosVol[x]=NormalizeDouble(StrToDouble(StringSubstr(s[x],start,length))*koeff_lot,DigitsMinLot(Symbols[x]));
//--- 
   start=end+1;// Type
//--- find position of ","
   end=StringFind(s[x],",",start);
//--- calculate the length of the substring
   length=end-start;
//--- get the Type
   SymPosType[x]=StrToInteger(StringSubstr(s[x],start,length));
//--- 
   start=end+1;
//--- find position of ","
   end=StringFind(s[x],",",start);
//--- calculate the length of the substring
   length=end-start;
//--- get Stop Loss
   SymPosSL[x]=NormalizeDouble(StrToDouble(StringSubstr(s[x],start,length)),digits(Symbols[x]));
//--- 
   start=end+1;
//--- find position of ","
   end=StringFind(s[x],",",start);
//--- calculate the length of the substring
   length=end-start;
//--- get Take Profit
   SymPosTP[x]=NormalizeDouble(StrToDouble(StringSubstr(s[x],start,length)),digits(Symbols[x]));
   return;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
